@@ -1,12 +1,12 @@
 # -*- coding: utf-8 -*-
 
-import time
-from xml.etree import ElementTree as etree
-import sys
-import re
+from xml.etree import cElementTree as etree
 import os
+import re
+import sys
+import time
 
-VERSION = 1.0
+VERSION = 1.1
 
 CONSTRAINTS = {
     'Time_Subdivision': 'Time subdivision of parent annotation\'s time interva'
@@ -17,6 +17,9 @@ CONSTRAINTS = {
     'Included_In': 'Time alignable annotations within the parent annotation\'s'
     ' time interval, gaps are allowed'}
 
+MIMES = {'wav': 'audio/x-wav', 'mpg': 'video/mpeg', 'mpeg': 'video/mpg',
+         'xml': 'text/xml'}
+
 
 class Eaf:
     """Read and write Elan's Eaf files.
@@ -27,66 +30,63 @@ class Eaf:
     :var dict licences: Licences included in the file.
     :var dict header: XML header.
     :var list media_descriptors: Linked files, where every file is of the
-                                 form: ``{attrib}``.
+        form: ``{attrib}``.
     :var list properties: Properties, where every property is of the form:
-                          ``(value, {attrib})``.
+        ``(value, {attrib})``.
     :var list linked_file_descriptors: Secondary linked files, where every
-                                       linked file is of the form:
-                                       ``{attrib}``.
-    :var dict timeslots: Timeslot data of the form:
-                         ``{TimslotID -> time(ms)}``.
-    :var dict tiers: Tier data of the form:
-                     ``{tier_name -> (aligned_annotations,
-                     reference_annotations, attributes, ordinal)}``,
+        linked file is of the form: ``{attrib}``.
+    :var dict timeslots: Timeslot data of the form: ``{id -> time(ms)}``.
+    :var dict tiers: Tiers, where every tier is of the form:
+        ``{tier_name -> (aligned_annotations, reference_annotations,
+        attributes, ordinal)}``,
 
-                     aligned_annotations of the form:
-                     ``[{annotation_id ->
-                     (begin_ts, end_ts, value, svg_ref)}]``,
+        aligned_annotations of the form: ``[{id -> (begin_ts, end_ts, value,
+        svg_ref)}]``,
 
-                     reference annotations of the form:
-                     ``[{annotation_id ->
-                     (reference, value, previous, svg_ref)}]``.
+        reference annotations of the form: ``[{id -> (reference, value,
+        previous, svg_ref)}]``.
     :var list linguistic_types: Linguistic types, where every type is of the
-                                form: ``{id -> attrib}``.
-    :var list locales: Locales, where every locale is of the form:
-                       ``{attrib}``.
-    :var dict constraints: Constraint data of the form:
-                           ``{stereotype -> description}``.
-    :var dict controlled_vocabularies: Controlled vocabulary data of the
-                                       form: ``{id ->
-                                       (descriptions, entries, ext_ref)}``,
+        form: ``{id -> attrib}``.
+    :var list locales: Locales, every locale is of the form: ``{attrib}``.
+    :var dict constraints: Constraints, every constraint is of the form:
+        ``{stereotype -> description}``.
+    :var dict controlled_vocabularies: Controlled vocabulary, where every
+        controlled vocabulary is of the form: ``{id -> (descriptions, entries,
+        ext_ref)}``,
 
-                                       descriptions of the form:
-                                       ``[(lang_ref, text)]``,
+        descriptions of the form: ``[(lang_ref, text)]``,
 
-                                       entries of the form:
-                                       ``{id -> (values, ext_ref)}``,
+        entries of the form: ``{id -> (values, ext_ref)}``,
 
-                                       values of the form:
-                                       ``[(lang_ref, description, text)]``.
+        values of the form:  ``[(lang_ref, description, text)]``.
     :var list external_refs: External references, where every reference is of
-                             the form ``[id, type, value]``.
+        the form ``[id, type, value]``.
     :var list lexicon_refs: Lexicon references, where every reference is of
-                            the form: ``[{attribs}]``.
+        the form: ``[{attribs}]``.
+
+    :var dict annotations: Dictionary of annotations of the form:
+        {id -> tier}``, this is only used internally.
     """
 
     def __init__(self, file_path=None, author='pympi'):
         """Construct either a new Eaf file or read on from a file/stream.
 
         :param str file_path: Path to read from, - for stdin. If ``None`` an
-                              empty Eaf file will be created.
+            empty Eaf file will be created.
         :param str author: Author of the file.
         """
-        self.naive_gen_ann, self.naive_gen_ts = False, False
+        self.maxts = None
+        self.maxaid = None
         self.annotation_document = {
             'AUTHOR': author,
-            'DATE': time.strftime("%Y-%m-%dT%H:%M:%S%z"),
+            'DATE': time.strftime('%Y-%m-%dT%H:%M:%S%z'),
             'VERSION': '2.8',
             'FORMAT': '2.8',
             'xmlns:xsi': 'http://www.w3.org/2001/XMLSchema-instance',
             'xsi:noNamespaceSchemaLocation':
                 'http://www.mpi.nl/tools/elan/EAFv2.8.xsd'}
         self.constraints = {}
+        self.annotations = {}
         self.controlled_vocabularies = {}
         self.header = {}
         self.licences = {}
@@ -102,7 +102,7 @@ class Eaf:
         self.new_time, self.new_ann = 0, 0
 
         if file_path is None:
-            self.add_linguistic_type('default-lt', None)
+            self.add_linguistic_type('default-lt')
             self.constraints = CONSTRAINTS.copy()
             self.properties.append(('0', {'NAME': 'lastUsedAnnotation'}))
             self.add_tier('default')
@@ -113,31 +113,33 @@ class Eaf:
         """Write the object to a file, if the file already exists a backup will
         be created with the ``.bak`` suffix.
 
-        :param str file_path: Path to write to, - for stdout.
-        :param bool pretty: Flag for pretty XML printing.
+        :param str file_path: Filepath to write to.
+        :param bool pretty: Flag for pretty XML printing (Only unset this if
+            you are afraid of wasting bytes because it won't print unneccesary
+            whitespace).
         """
         to_eaf(file_path, self, pretty)
 
-    def to_textgrid(self, excluded_tiers=[], included_tiers=[]):
+    def to_textgrid(self, filtin=[], filtex=[], regex=False):
         """Convert the object to a :class:`pympi.Praat.TextGrid` object.
 
-        :param list excluded_tiers: Specifically exclude these tiers.
-        :param list included_tiers: Only include this tiers, when empty all are
-                                    included.
+        :param list filtin: Include only tiers in this list, if empty
+            all tiers are included.
+        :param list filtex: Exclude all tiers in this list.
+        :param bool regex: If this flag is set the filters are seen as regexes.
         :returns: :class:`pympi.Praat.TextGrid` representation.
         :raises ImportError: If the pympi.Praat module can't be loaded.
         """
         from pympi.Praat import TextGrid
         _, end = self.get_full_time_interval()
         tgout = TextGrid(xmax=end)
-        tiers = [a for a in self.tiers if a not in excluded_tiers]
-        if included_tiers:
-            tiers = [a for a in tiers if a in included_tiers]
-        for tier in tiers:
+        func = (lambda x, y: re.match(x, y)) if regex else lambda x, y: x == y
+        for tier in self.tiers:
+            if (filtin and not any(func(f, tier) for f in filtin)) or\
+                    (filtex and any(func(f, tier) for f in filtex)):
+                continue
             ctier = tgout.add_tier(tier)
             for intv in self.get_annotation_data_for_tier(tier):
-                if intv[0] == intv[1]:
-                    continue
                 ctier.add_interval(intv[0]/1000.0, intv[1]/1000.0, intv[2])
         return tgout
 
@@ -150,20 +152,42 @@ class Eaf:
         """
         from copy import deepcopy
         eaf_out = deepcopy(self)
-        for tier in eaf_out.tiers.itervalues():
-            rems = []
-            for ann in tier[0]:
-                if eaf_out.timeslots[tier[0][ann][1]] > end or\
-                        eaf_out.timeslots[tier[0][ann][0]] < start:
-                    rems.append(ann)
-            for r in rems:
-                del tier[0][r]
+        for t in eaf_out.get_tier_names():
+            for ab, ae, value in eaf_out.get_annotation_data_for_tier(t):
+                if ab > end or ae < start:
+                    eaf_out.remove_annotation(t, (start-end)/2, False)
+        eaf_out.clean_time_slots()
         return eaf_out
 
     def get_linked_files(self):
         """Give all linked files."""
         return self.media_descriptors
 
+    def get_secondary_linked_files(self):
+        """Give all linked files."""
+        return self.linked_file_descriptors
+
+    def add_secondary_linked_file(self, file_path, relpath=None, mimetype=None,
+                                  time_origin=None, assoc_with=None):
+        """Add a secondary linked file.
+
+        :param str file_path: Path of the file.
+        :param str relpath: Relative path of the file.
+        :param str mimetype: Mimetype of the file, if ``None`` it tries to
+            guess it according to the file extension which currently only works
+            for wav, mpg, mpeg and xml.
+        :param int time_origin: Time origin for the media file.
+        :param str assoc_with: Associated with field.
+        :raises KeyError: If mimetype had to be guessed and a non standard
+                          extension or an unknown mimetype.
+        """
+        if mimetype is None:
+            mimetype = MIMES[file_path.split('.')[-1]]
+        self.linked_file_descriptors.append({
+            'LINK_URL': file_path, 'RELATIVE_LINK_URL': relpath,
+            'MIME_TYPE': mimetype, 'TIME_ORIGIN': time_origin,
+            'ASSOCIATED_WITH': assoc_with})
+
     def add_linked_file(self, file_path, relpath=None, mimetype=None,
                         time_origin=None, ex_from=None):
         """Add a linked file.
@@ -171,17 +195,15 @@ class Eaf:
         :param str file_path: Path of the file.
         :param str relpath: Relative path of the file.
         :param str mimetype: Mimetype of the file, if ``None`` it tries to
-                             guess it according to the file extension which
-                             currently only works for wav, mpg, mpeg and xml.
+            guess it according to the file extension which currently only works
+            for wav, mpg, mpeg and xml.
         :param int time_origin: Time origin for the media file.
         :param str ex_from: Extracted from field.
         :raises KeyError: If mimetype had to be guessed and a non standard
                           extension or an unknown mimetype.
         """
         if mimetype is None:
-            mimes = {'wav': 'audio/x-wav', 'mpg': 'video/mpeg',
-                     'mpeg': 'video/mpg', 'xml': 'text/xml'}
-            mimetype = mimes[file_path.split('.')[-1]]
+            mimetype = MIMES[file_path.split('.')[-1]]
         self.media_descriptors.append({
             'MEDIA_URL': file_path, 'RELATIVE_MEDIA_URL': relpath,
             'MIME_TYPE': mimetype, 'TIME_ORIGIN': time_origin,
@@ -318,11 +340,13 @@ class Eaf:
         :returns: List of annotations within that time.
         :raises KeyError: If the tier is non existent.
         """
-        anns = self.tiers[id_tier][0]
-        return sorted([
-            (self.timeslots[m[0]], self.timeslots[m[1]], m[2])
-            for m in anns.itervalues() if self.timeslots[m[1]] >= start and
-            self.timeslots[m[0]] <= end])
+        anns = ((self.timeslots[a[0]], self.timeslots[a[1]], a[2])
+                for a in self.tiers[id_tier][0].itervalues())
+        return sorted(a for a in anns if a[1] >= start and a[0] <= end)
+#        return sorted([
+#            (self.timeslots[m[0]], self.timeslots[m[1]], m[2])
+#            for m in anns.itervalues() if self.timeslots[m[1]] >= start and
+#            self.timeslots[m[0]] <= end])
 
     def remove_all_annotations_from_tier(self, id_tier, clean=True):
         """remove all annotations from a tier
@@ -330,6 +354,8 @@ class Eaf:
         :param str id_tier: Name of the tier.
         :raises KeyError: If the tier is non existent.
         """
+        for aid in self.tiers[id_tier][0].keys()+self.tiers[id_tier][1].keys():
+            del(self.annotations[aid])
         self.tiers[id_tier][0].clear()
         self.tiers[id_tier][1].clear()
         if clean:
@@ -358,8 +384,9 @@ class Eaf:
             raise ValueError('Start is negative...')
         start_ts = self.generate_ts_id(start)
         end_ts = self.generate_ts_id(end)
-        self.tiers[id_tier][0][self.generate_annotation_id()] =\
-            (start_ts, end_ts, value, svg_ref)
+        aid = self.generate_annotation_id()
+        self.annotations[aid] = id_tier
+        self.tiers[id_tier][0][aid] = (start_ts, end_ts, value, svg_ref)
 
     def remove_annotation(self, id_tier, time, clean=True):
         """Remove an annotation in a tier, if you need speed the best thing is
@@ -377,38 +404,70 @@ class Eaf:
                   self.timeslots[a[1][0]] <= time and
                   self.timeslots[a[1][1]] >= time]:
             del(self.tiers[id_tier][0][b[0]])
+            del(self.annotations[b[0]])
             removed += 1
         if clean:
             self.clean_time_slots()
         return removed
 
-    def insert_ref_annotation(self, id_tier, ref, value, prev=None,
-                              svg_ref=None):
+    def insert_ref_annotation(self, id_tier, tier2, time, value,
+                              prev=None, svg=None):
         """Insert a reference annotation.
-        NOTE THAT THIS IS NOT VERY USABLE BECAUSE YOU DON'T KNOW THE INTERNAL
-        ANNOTATION ID'S
 
         :param str id_tier: Name of the tier.
-        :param str ref: Id of the referenced annotation.
+        :param str tier2: Tier of the referenced annotation.
+        :param int time: Time of the referenced annotation.
         :param str value: Value of the annotation.
         :param str prev: Id of the previous annotation.
         :param str svg_ref: Svg reference.
         :raises KeyError: If the tier is non existent.
-        :raises ValueError: If the tier already contains normal annotations.
+        :raises ValueError: If the tier already contains normal annotations or
+            if there is no annotation in the tier on the time to reference to.
         """
         if self.tiers[id_tier][0]:
             raise ValueError('This tier already contains normal annotations.')
-        self.tiers[id_tier][1][self.generate_annotation_id()] =\
-            (ref, value, prev, svg_ref)
+        ann = None
+        for aid, (begin, end, value, _) in self.tiers[tier2][0].iteritems():
+            begin = self.timeslots[begin]
+            end = self.timeslots[end]
+            if begin <= time and end >= time:
+                ann = aid
+                break
+        if not ann:
+            raise ValueError('There is no annotation to reference to.')
+        aid = self.generate_annotation_id()
+        self.annotations[aid] = id_tier
+        self.tiers[id_tier][1][aid] = (ann, value, prev, svg)
 
     def get_ref_annotation_data_for_tier(self, id_tier):
         """"Give a list of all reference annotations of the form:
-        ``[{id -> (ref, value, previous, svg_ref}]``
+        ``[(start, end, value, refvalue)]``
 
         :param str id_tier: Name of the tier.
         :raises KeyError: If the tier is non existent.
+        :yields: Reference annotations within that tier.
         """
-        return self.tiers[id_tier][1]
+        for aid, (ref, value, prev, _) in self.tiers[id_tier][1]:
+            refann = self.tiers[self.annotations[ref]][0][ref]
+            yield (self.timeslots[refann[0]], self.timeslots[refann[1]],
+                   value, refann[2])
+
+    def get_ref_annotation_at_time(self, tier, time):
+        """Give the ref annotations at the given time.
+
+        :param str tier: Name of the tier.
+        :param int time: Time of the annotation of the parent.
+        :returns: List of annotations at that time.
+        :raises KeyError: If the tier is non existent.
+        """
+        bucket = []
+        for aid, (ref, _, _, _) in self.tiers[tier][1].iteritems():
+            begin, end, value, _ = self.tiers[self.annotations[ref]][0][ref]
+            begin = self.timeslots[begin]
+            end = self.timeslots[end]
+            if begin <= time and end >= time:
+                bucket.append((begin, end, value))
+        return bucket
 
     def remove_controlled_vocabulary(self, cv):
         """Remove a controlled vocabulary.
@@ -422,22 +481,14 @@ class Eaf:
         """Generate the next annotation id, this function is mainly used
         internally.
         """
-        if self.naive_gen_ann:
-            new = self.last_ann+1
-            self.last_ann = new
+        if self.maxaid is None:
+            self.maxaid = 1
+            valid_anns = [a for a in self.annotations if re.match('.\d+', a)]
+            if valid_anns:
+                self.maxaid = max(int(x[1:]) for x in valid_anns)+1
         else:
-            new = 1
-            an = [a for y in self.tiers.itervalues() for x in y[:2] for a in x]
-            anns = [int(x[1:]) for x in an if re.match('.\d+', x)]
-            if len(anns) > 0:
-                newann = set(xrange(1, max(anns))).difference(anns)
-                if len(newann) == 0:
-                    new = max(anns)+1
-                    self.naive_gen_ann = True
-                    self.last_ann = new
-                else:
-                    new = sorted(newann)[0]
-        return 'a{:d}'.format(new)
+            self.maxaid += 1
+        return 'a{:d}'.format(self.maxaid)
 
     def generate_ts_id(self, time=None):
         """Generate the next timeslot id, this function is mainly used
@@ -448,40 +499,31 @@ class Eaf:
         """
         if time and time < 0:
             raise ValueError('Time is negative...')
-        if self.naive_gen_ts:
-            new = self.last_ts+1
-            self.last_ts = new
+        if self.maxts is None:
+            self.maxts = 1
+            valid_ts = [a for a in self.timeslots if re.match('..\d+', a)]
+            if valid_ts:
+                self.maxts = max(int(x[2:]) for x in valid_ts)+1
         else:
-            new = 1
-            tss = {int(x[2:]) for x in self.timeslots if re.match('..\d+', x)}
-            if len(tss) > 0:
-                newts = set(xrange(1, max(tss))).difference(tss)
-                if len(newts) == 0:
-                    new = max(tss)+1
-                    self.naive_gen_ts = True
-                    self.last_ts = new
-                else:
-                    new = sorted(newts)[0]
-        ts = 'ts{:d}'.format(new)
+            self.maxts += 1
+        ts = 'ts{:d}'.format(self.maxts)
         self.timeslots[ts] = time
         return ts
 
     def clean_time_slots(self):
         """Clean up all unused timeslots.
         .. warning:: This can and will take time for larger tiers. When you
-           want to do a lot of operations on a lot of tiers please unset the
-           flags for cleaning in the functions so that the cleaning is only
-           performed afterwards.
+        want to do a lot of operations on a lot of tiers please unset the flags
+        for cleaning in the functions so that the cleaning is only performed
+        afterwards.
         """
-        ts_in_tier = set(sum([a[0:2] for tier in self.tiers.itervalues()
-                              for a in tier[0].itervalues()], ()))
-        ts_avail = set(self.timeslots)
-        for a in ts_in_tier.symmetric_difference(ts_avail):
+        ts_in_tier = ((a[0], a[1]) for tier in self.tiers.itervalues() for a in
+                      tier[0].itervalues())
+        ts_in_tier = {a for b in ts_in_tier for a in b}
+        for a in ts_in_tier.symmetric_difference(self.timeslots):
             del(self.timeslots[a])
-        self.naive_gen_ts = False
-        self.naive_gen_ann = False
 
-    def merge_tiers(self, tiers, tiernew=None, gapt=0):
+    def merge_tiers(self, tiers, tiernew=None, gapt=0, sep='_', safe=False):
         """Merge tiers into a new tier and when the gap is lower then the
         threshhold glue the annotations together.
 
@@ -490,28 +532,30 @@ class Eaf:
                             generated.
         :param int gapt: Threshhold for the gaps, if the this is set to 10 it
                          means that all gaps below 10 are ignored.
+        :param str sep: Separator for the merged annotations.
+        :param bool safe: Ignore zero length annotations(when working with
+            possible malformed data).
         :raises KeyError: If a tier is non existent.
         """
         if tiernew is None:
             tiernew = '{}_merged'.format('_'.join(tiers))
-        if tiernew in self.tiers:
-            self.remove_tier(tiernew)
         self.add_tier(tiernew)
-        ad = map(sorted, map(self.get_annotation_data_for_tier, tiers))
-        while any(ad):
-            current = []
-            queue = [(a[0], i) for i, a in enumerate(ad) if a]
-            while queue:
-                queue.sort()
-                a, i = queue.pop(0)
-                if not current or a[0] < max(x[1] for x in current)+gapt:
-                    current.append(ad[i].pop(ad[i].index(a)))
-                    if ad[i]:
-                        queue.append((ad[i][0], i))
-            start = min(x[0] for x in current)
-            end = max(x[1] for x in current)
-            self.insert_annotation(tiernew, start, end,
-                                   '_'.join(x[2] for x in current))
+        aa = [(sys.maxint, sys.maxint, None)] + sorted((
+            a for t in tiers for a in self.get_annotation_data_for_tier(t)),
+            reverse=True)
+        l = None
+        while aa:
+            begin, end, value = aa.pop()
+            if l is None:
+                l = [begin, end, [value]]
+            elif begin - l[1] >= gapt:
+                if not safe or l[1] > l[0]:
+                    self.insert_annotation(tiernew, l[0], l[1], sep.join(l[2]))
+                l = [begin, end, [value]]
+            else:
+                if end > l[1]:
+                    l[1] = end
+                l[2].append(value)
 
     def shift_annotations(self, time):
         """Shift all annotations in time. Annotations that are in the beginning
@@ -538,37 +582,40 @@ class Eaf:
             for name, aid in squashed:
                 start, end, value, _ = self.tiers[name][0][aid]
                 del(self.tiers[name][0][aid])
+                del(self.annotations[aid])
                 total_re.append(
                     (name, self.timeslots[start], self.timeslots[end], value))
         return total_sq, total_re
 
     def filter_annotations(self, tier, tier_name=None, filtin=None,
-                           filtex=None, regex=False):
-        """Filter annotations in a tier
+                           filtex=None, regex=False, safe=False):
+        """Filter annotations in a tier using an exclusive and/or inclusive
+        filter.
 
-        :param str tier: Name of the tier:
-        :param str tier_name: Name of the new tier, when ``None`` the name will
-                              be generated.
+        :param str tier: Name of the tier.
+        :param str tier_name: Name of the output tier, when ``None`` the name
+            will be generated.
         :param list filtin: List of strings to be included, if None all
-                            annotations all is included.
+            annotations all is included.
         :param list filtex: List of strings to be excluded, if None no strings
-                            are excluded.
+            are excluded.
         :param bool regex: If this flag is set, the filters are seen as regex
-                           matches.
+            matches.
+        :param bool safe: Ignore zero length annotations(when working with
+            possible malformed data).
         :raises KeyError: If the tier is non existent.
         """
         if tier_name is None:
             tier_name = '{}_filter'.format(tier)
-        if tier_name in self.tiers:
-            self.remove_tier(tier_name)
         self.add_tier(tier_name)
-        for b, e, v in self.get_annotation_data_for_tier(tier):
-            inc = True if not filtin else\
-                any(re.match(f, v) if regex else v == f for f in filtin)
-            exc = True if not filtex else\
-                all(not re.match(f, v) if regex else v != f for f in filtex)
-            if inc and exc:
-                self.insert_annotation(tier_name, b, e, v)
+        func = (lambda x, y: re.match(x, y)) if regex else lambda x, y: x == y
+        for begin, end, value in self.get_annotation_data_for_tier(tier):
+            if (filtin and not any(func(f, value) for f in filtin)) or\
+                    (filtex and any(func(f, value) for f in filtex)):
+                continue
+            if not safe or end > begin:
+                self.insert_annotation(tier_name, begin, end, value)
+        self.clean_time_slots()
 
     def get_full_time_interval(self):
         """Give the full time interval of the file.
@@ -580,9 +627,9 @@ class Eaf:
              max(self.timeslots.itervalues()))
 
     def create_gaps_and_overlaps_tier(self, tier1, tier2, tier_name=None,
-                                      maxlen=-1):
+                                      maxlen=-1, fast=False):
         """Create a tier with the gaps and overlaps of the annotations.
-        For types see :func:`get_gaps_and_overlaps_duration`
+        For types see :func:`get_gaps_and_overlaps`
 
         :param str tier1: Name of the first tier.
         :param str tier2: Name of the second tier.
@@ -590,6 +637,7 @@ class Eaf:
                               be generated.
         :param int maxlen: Maximum length of gaps (skip longer ones), if ``-1``
                            no maximum will be used.
+        :param bool fast: Flag for using the fast method.
         :returns: List of gaps and overlaps of the form:
                   ``[(type, start, end)]``.
         :raises KeyError: If a tier is non existent.
@@ -597,20 +645,23 @@ class Eaf:
         """
         if tier_name is None:
             tier_name = '{}_{}_ftos'.format(tier1, tier2)
-        if tier_name in self.tiers:
-            self.remove_tier(tier_name)
         self.add_tier(tier_name)
         ftos = []
-        for fto in self.get_gaps_and_overlaps(tier1, tier2, maxlen):
+        ftogen = self.get_gaps_and_overlaps2(tier1, tier2, maxlen) if fast else\
+            self.get_gaps_and_overlaps(tier1, tier2, maxlen)
+        for fto in ftogen:
             ftos.append(fto)
             if fto[1]-fto[0] >= 1:
                 self.insert_annotation(tier_name, fto[0], fto[1], fto[2])
+        self.clean_time_slots()
         return ftos
 
     def get_gaps_and_overlaps(self, tier1, tier2, maxlen=-1):
         """Give gaps and overlaps. The return types are shown in the table
         below. The string will be of the format: ``id_tiername_tiername``.
 
+        .. note:: There is also a faster method: :func:`get_gaps_and_overlaps2`
+
         For example when a gap occurs between tier1 and tier2 and they are
         called ``speakerA`` and ``speakerB`` the annotation value of that gap
         will be ``G12_speakerA_speakerB``.
@@ -618,8 +669,8 @@ class Eaf:
         | The gaps and overlaps are calculated using Heldner and Edlunds
           method found in:
         | *Heldner, M., & Edlund, J. (2010). Pauses, gaps and overlaps in
-         conversations. Journal of Phonetics, 38(4), 555–568.
-         doi:10.1016/j.wocn.2010.08.002*
+          conversations. Journal of Phonetics, 38(4), 555–568.
+          doi:10.1016/j.wocn.2010.08.002*
 
         +-----+---------------------------------------------+
         | id  | Description                                 |
@@ -704,11 +755,39 @@ class Eaf:
                         ('W21', tier2, tier1)
                     yield (line1[i][1], line1[i][2]-1, '_'.join(t))
 
+    def get_gaps_and_overlaps2(self, tier1, tier2, maxlen=-1):
+        """Faster variant of :func:`get_gaps_and_overlaps`.
+
+        :param str tier1: Name of the first tier.
+        :param str tier2: Name of the second tier.
+        :param int maxlen: Maximum length of gaps (skip longer ones), if ``-1``
+                           no maximum will be used.
+        :yields: Tuples of the form ``[(start, end, type)]``.
+        :raises KeyError: If a tier is non existent.
+        """
+        ad = sorted(((a, i+1) for i, t in enumerate([tier1, tier2]) for a in
+                     self.get_annotation_data_for_tier(t)), reverse=True)
+        if ad:
+            last = (lambda x: (x[0][0], x[0][1], x[1]))(ad.pop())
+            thr = lambda x, y: maxlen == -1 or abs(x-y) < maxlen
+            while ad:
+                (begin, end, _), current = ad.pop()
+                if last[2] == current and thr(begin, last[1]):
+                    yield (last[1], begin, 'P{}'.format(current))
+                elif last[0] < begin and last[1] > end:
+                    yield (begin, end, 'W{}{}'.format(last[2], current))
+                    continue
+                elif last[1] > begin:
+                    yield (begin, last[1], 'O{}{}'.format(last[2], current))
+                elif last[1] < begin and thr(begin, last[1]):
+                    yield (last[1], begin, 'G{}{}'.format(last[2], current))
+                last = (begin, end, current)
+
     def create_controlled_vocabulary(self, cv_id, descriptions, entries,
                                      ext_ref=None):
         """Create a controlled vocabulary.
         .. warning:: This is a very raw implementation and you should check the
-                     Eaf file format specification for the entries.
+        Eaf file format specification for the entries.
 
         :param str cv_id: Name of the controlled vocabulary.
         :param list descriptions: List of descriptions.
@@ -784,6 +863,59 @@ class Eaf:
         """
         return self.linguistic_types[lingtype]
 
+    def remove_linked_files(self, file_path=None, relpath=None, mimetype=None,
+                            time_origin=None, ex_from=None):
+        """Remove all linked files that match all the criteria, criterias that
+        are ``None`` are ignored.
+
+        :param str file_path: Path of the file.
+        :param str relpath: Relative filepath.
+        :param str mimetype: Mimetype of the file.
+        :param int time_origin: Time origin.
+        :param str ex_from: Extracted from.
+        """
+        for attrib in self.media_descriptors[:]:
+            if file_path is not None and attrib['MEDIA_URL'] != file_path:
+                continue
+            if relpath is not None and attrib['RELATIVE_MEDIA_URL'] != relpath:
+                continue
+            if mimetype is not None and attrib['MIME_TYPE'] != mimetype:
+                continue
+            if time_origin is not None and\
+                    attrib['TIME_ORIGIN'] != time_origin:
+                continue
+            if ex_from is not None and attrib['EXTRACTED_FROM'] != ex_from:
+                continue
+            del(self.media_descriptors[self.media_descriptors.index(attrib)])
+
+    def remove_secondary_linked_files(self, file_path=None, relpath=None,
+                                      mimetype=None, time_origin=None,
+                                      assoc_with=None):
+        """Remove all secondary linked files that match all the criteria,
+        criterias that are ``None`` are ignored.
+
+        :param str file_path: Path of the file.
+        :param str relpath: Relative filepath.
+        :param str mimetype: Mimetype of the file.
+        :param int time_origin: Time origin.
+        :param str ex_from: Extracted from.
+        """
+        for attrib in self.linked_file_descriptors[:]:
+            if file_path is not None and attrib['LINK_URL'] != file_path:
+                continue
+            if relpath is not None and attrib['RELATIVE_LINK_URL'] != relpath:
+                continue
+            if mimetype is not None and attrib['MIME_TYPE'] != mimetype:
+                continue
+            if time_origin is not None and\
+                    attrib['TIME_ORIGIN'] != time_origin:
+                continue
+            if assoc_with is not None and\
+                    attrib['ASSOCIATED_WITH'] != assoc_with:
+                continue
+            del(self.linked_file_descriptors[
+                self.linked_file_descriptors.index(attrib)])
+
 
 def parse_eaf(file_path, eaf_obj):
     """Parse an EAF file
@@ -842,6 +974,7 @@ def parse_eaf(file_path, eaf_obj):
                                                '' if not list(elem2)[0].text
                                                else list(elem2)[0].text,
                                                svg_ref)
+                            eaf_obj.annotations[annot_id] = tier_id
                         elif elem2.tag == 'REF_ANNOTATION':
                             annotRef = elem2.attrib['ANNOTATION_REF']
                             previous = elem2.attrib.get('PREVIOUS_ANNOTATION',
@@ -855,6 +988,7 @@ def parse_eaf(file_path, eaf_obj):
                                             '' if not list(elem2)[0].text else
                                             list(elem2)[0].text,
                                             previous, svg_ref)
+                            eaf_obj.annotations[annotId] = tier_id
             eaf_obj.tiers[tier_id] = (align, ref, elem.attrib, tier_number)
             tier_number += 1
         # Linguistic type
@@ -1023,35 +1157,3 @@ def to_eaf(file_path, eaf_obj, pretty=True):
         os.rename(file_path, '{}.bak'.format(file_path))
     etree.ElementTree(ANNOTATION_DOCUMENT).write(
         file_path, xml_declaration=True, encoding='UTF-8')
-
-# def eaf_from_chat(filepath):
-#     prebegin = []
-#     data = {}
-#     anndata = []
-#     a = Eaf()
-#     with open(filepath, 'r') as fin:
-#         line = fin.readline().strip()
-#         while line != '@Begin':
-#             prebegin.append(line)
-#             line = fin.readline().strip()
-#         line = fin.readline().strip()
-#         while line.startswith('@'):
-#             split = map(lambda x: x.strip(), line.split(':'))
-#             if split[0].strip() == '@Media':
-#                 split = split[1].split(',')
-#                 source = split[0]
-#                 if split[1] == 'audio':
-#                     a.add_linked_file(
-#                         '{}/{}.wav'.format(os.getcwd(), split[0]),
-#                         './{}.wav'.format(split[0]))
-#             elif split
-#             line = fin.readline().strip()
-#         while line != '@End':
-#             splits = line.split('\x15')
-#             if len(splits) == 3:
-#                 value, time, _ = line.split('\x15')
-#                 start, end = map(int, time.split('_'))
-#                 tier, value = value.split('\t')
-#                 tier = tier[1:-1]
-#                 anndata.append((tier, start, end, value))
-#             line = fin.readline().strip()
